/* 
 **  mod_memcached.c -- Apache module for put/delete/head/get
 */ 

#include "httpd.h"
#include "apr.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "version.h"
#include <libmemcached/memcached.h>

#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>

#include <assert.h>

module AP_MODULE_DECLARE_DATA memcached_module;

typedef struct {
  int delete;
  int put;
  int get;
  char *servers;
} memcached_conf;

void
set_config_defaults(apr_pool_t *p, memcached_conf *cfg)
{
  bzero(cfg, sizeof(memcached_conf));
}

static void *create_dir_mconfig(apr_pool_t *p, char *dir) 
{
  memcached_conf *cfg;

  cfg= apr_pcalloc(p, sizeof(memcached_conf));
  set_config_defaults(p, cfg);

  return (void *) cfg;
}

static void *merge_dir_mconfig(apr_pool_t *p, void *origin, void *new) 
{
  memcached_conf *cfg_origin= (memcached_conf *)origin;
  memcached_conf *cfg_new= (memcached_conf *)new;

  return (void *) ( cfg_new ? cfg_new : cfg_origin );
}

int location(request_rec *r, char **filename)
{
  char *key;

  key= (char *)rindex(r->uri, '/');

  if (key)
  {
    key++;

    if (*key)
    {
      *filename= key;
      return HTTP_OK;
    }
    else
    {
      *filename= NULL;
      return HTTP_NOT_FOUND;
    }
  }
  else
  {
    *filename= NULL;
    return HTTP_OK;
  }
}

int get_method(request_rec *r, memcached_st *memc)
{
  memcached_return rc;
  char *key;
  char *value;
  size_t value_length;
  uint32_t flags;
  int result;

  if ((result = ap_discard_request_body(r)) != OK)
    return result;

  result= location(r, &key);

  if (r->header_only)
    return(result);

  if (result != HTTP_OK)
    return(result);

  value= memcached_get(memc, key, strlen(key), &value_length, &flags, &rc);

  if (value == NULL)
    result= HTTP_NOT_FOUND;
  else
  {
    request_rec *subr;

    subr = (request_rec *) ap_sub_req_lookup_file(key, r, NULL);
    printf("Content type %s\n", subr->content_type);
    ap_set_content_type(r, subr->content_type);
    r->content_type = apr_pstrdup(r->pool, subr->content_type);
    ap_destroy_sub_req(subr);

    result= OK;
    ap_rwrite(value, value_length, r);
  }

  free(value);

  return result;
}

int delete_method(request_rec *r, memcached_st *memc)
{
  memcached_return rc;
  char *key;
  int result;

  memc= memcached_create(NULL);

  if ((result = ap_discard_request_body(r)) != OK)
    return result;

  result= location(r, &key);

  if (result != HTTP_OK)
    return(result);

  rc= memcached_delete(memc, key, strlen(key), 0);

  if (rc == MEMCACHED_SUCCESS)
    result= HTTP_OK;
  else
    result= HTTP_NOT_FOUND;

  return result;
}

int put_method(request_rec *r, memcached_st *memc)
{
  memcached_return rc;
  char *buffer= NULL;
  size_t buffer_length= 0;
  char *key;
  int rv, seen_eos;
  apr_bucket_brigade *bb;


  if (location(r, &key) != HTTP_NOT_FOUND)
    return HTTP_CONFLICT;

  bb= apr_brigade_create(r->pool, r->connection->bucket_alloc);
  seen_eos = 0;

  do {
    apr_bucket *bucket;

    rv= ap_get_brigade(r->input_filters, bb, AP_MODE_READBYTES,
                       APR_BLOCK_READ, HUGE_STRING_LEN);

    if (rv != APR_SUCCESS) 
    {
      if (rv == AP_FILTER_ERROR) 
      {
        apr_brigade_destroy(bb);
        return rv;
      }
      else 
      {
        apr_brigade_destroy(bb);
        return HTTP_BAD_REQUEST;
      }
    }

    for (bucket= APR_BRIGADE_FIRST(bb);
         bucket != APR_BRIGADE_SENTINEL(bb);
         bucket= APR_BUCKET_NEXT(bucket)) 
    {
      const char *data;
      apr_size_t len;

      if (APR_BUCKET_IS_EOS(bucket)) 
      {
        seen_eos = 1;
        break;
      }

      if (bucket->length == 0)
        continue;

      rv= apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ);
      if (rv != APR_SUCCESS) 
      {
        apr_brigade_destroy(bb);
        return HTTP_BAD_REQUEST;
      }

      buffer= apr_pstrcat(r->pool, buffer, NULL);
      buffer_length+= len;
    }
    apr_brigade_cleanup(bb);
  } while (!seen_eos);

  rc= memcached_set(memc, key, strlen(key), 
                    buffer, buffer_length,
                    (time_t)0, (uint32_t)0);


  return OK;
}

/* The sample content handler */
static int memcached_handler(request_rec *r)
{
  memcached_conf *cfg= ap_get_module_config (r->per_dir_config, &memcached_module);
  memcached_st *memc;
  memcached_server_st *servers;
  memcached_return rc;
  int result;

  apr_table_setn(r->headers_out, "X-Powered-By", "ModMemcached/"VERSION);

  memc= memcached_create(NULL);
  servers= memcached_servers_parse(cfg->servers);
  rc= memcached_server_push(memc, servers);

  switch (r->method_number)
  {
  case M_DELETE:
    if (cfg->delete == 0)
      result= DECLINED;
    else /* We discard the body here so that we can resuse delete_method for update */
    {
      ap_discard_request_body(r);
      result= delete_method(r, memc);
    }
    break;
  case M_PUT:
    if (cfg->put == 0)
      result= DECLINED;
    else
      result= put_method(r, memc);
    break;
  case M_GET:
    if (cfg->get == 0)
      result= DECLINED;
    else
      result= get_method(r, memc);
    break;
  default:
    result= HTTP_METHOD_NOT_ALLOWED;
    break;
  }

  memcached_server_list_free(servers);
  memcached_free(memc);

  return result;
}

static void memcached_register_hooks(apr_pool_t *p)
{
  ap_hook_handler(memcached_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

static const command_rec memcached_cmds[]= {
  AP_INIT_FLAG("Memcached", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, get), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_FLAG("MemcachedDelete", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, delete), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_FLAG("MemcachedPut", ap_set_flag_slot, (void *) APR_OFFSETOF(memcached_conf, put), OR_ALL, 
               "On or Off, controls if this module is enabled or not (default off)"),
  AP_INIT_TAKE1("MemcachedServers", ap_set_string_slot, (void *) APR_OFFSETOF(memcached_conf, servers), OR_ALL, 
                "List of memcached servers to use (in server:port format)"),
  {NULL}
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA memcached_module = {
  STANDARD20_MODULE_STUFF, 
  create_dir_mconfig,    /* create per-dir    config structures */
  merge_dir_mconfig,     /* merge  per-dir    config structures */
  NULL,                  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  memcached_cmds,                  /* table of config file commands       */
  memcached_register_hooks  /* register hooks                      */
};

